( 3iPianoLights.fs - AE - 2021-05-10---Mon---00:12 )
( This program is for the 32-key LED piano lights board for teaching music theory. )
( The lights board is run from the Arduino Nano running 3iForth.avr code, and run through GForth )
(     running 3iForth.fs, 3iSound.fs, 3iShift.fs )

\ USAGE:
\ include 3iForth.fs
\ 1 open-com
\ include 3iSound.fs
\ include 3iShift.fs
\ init-shiftout


\ mapping table, left to right, white and black keys intermixed
create keymap 
\ keys 1-8
  $80 c, 0 c, 0 c, 0 c, 
  $40 c, 0 c, 0 c, 0 c, 
  $20 c, 0 c, 0 c, 0 c, 
  $10 c, 0 c, 0 c, 0 c, 
  $08 c, 0 c, 0 c, 0 c, 
  $04 c, 0 c, 0 c, 0 c, 
  $02 c, 0 c, 0 c, 0 c, 
  $01 c, 0 c, 0 c, 0 c, 
\ keys 9-16
  0 c, $80 c, 0 c, 0 c,
  0 c, $40 c, 0 c, 0 c,
  0 c, $20 c, 0 c, 0 c,
  0 c, $10 c, 0 c, 0 c,
  0 c, $08 c, 0 c, 0 c,
  0 c, $04 c, 0 c, 0 c,
  0 c, $02 c, 0 c, 0 c,
  0 c, $01 c, 0 c, 0 c,
\ keys 17-24
  0 c, 0 c, $80 c, 0 c,
  0 c, 0 c, $40 c, 0 c,
  0 c, 0 c, $20 c, 0 c,
  0 c, 0 c, $10 c, 0 c,
  0 c, 0 c, $08 c, 0 c,
  0 c, 0 c, $04 c, 0 c,
  0 c, 0 c, $02 c, 0 c,
  0 c, 0 c, $01 c, 0 c,
\ keys 25-32
  0 c, 0 c, 0 c, $80 c,
  0 c, 0 c, 0 c, $40 c,
  0 c, 0 c, 0 c, $20 c,
  0 c, 0 c, 0 c, $10 c,
  0 c, 0 c, 0 c, $08 c,
  0 c, 0 c, 0 c, $04 c,
  0 c, 0 c, 0 c, $02 c,
  0 c, 0 c, 0 c, $01 c,

: get-key ( keyid -- ) 1 - 4 *  keymap + 
    c@+ swap c@+ swap c@+ swap c@ ;
: light-key ( keyid -- ) get-key shout32 ;


create keycode  \ integer values (4 bytes) for each key
  1 31 lshift ,
  1 30 lshift ,
  1 29 lshift ,
  1 28 lshift ,
  1 27 lshift ,
  1 26 lshift ,
  1 25 lshift ,
  1 24 lshift ,
  1 23 lshift ,
  1 22 lshift ,
  1 21 lshift ,
  1 20 lshift ,
  1 19 lshift ,
  1 18 lshift ,
  1 17 lshift ,
  1 16 lshift ,
  1 15 lshift ,
  1 14 lshift ,
  1 13 lshift ,
  1 12 lshift ,
  1 11 lshift ,
  1 10 lshift ,
  1  9 lshift ,
  1  8 lshift ,
  1  7 lshift ,
  1  6 lshift ,
  1  5 lshift ,
  1  4 lshift ,
  1  3 lshift ,
  1  2 lshift ,
  1  1 lshift ,
  1 ,
  
: get-code ( key-id -- ) 1 - 4 * keycode + @ ;

$FF 24 lshift constant maskb4
$FF 16 lshift constant maskb3
$FF 8 lshift  constant maskb2
$FF           constant maskb1

\ ability to light any array of keys at the same time
: add-codes ( key1 ... keyn n -- ) >r get-code r> 1 - 0 do swap get-code + loop ;
: int2bytes ( int -- byte-msb byte3 byte2 byte-lsb ) \ extract the 4 bytes, lsb top of stack
    >r r@ 24 rshift
    r@ maskb3 and 16 rshift
    r@ maskb2 and 8 rshift
    r> maskb1 and 
    ;
: lightn ( key1 ... keyn n -- )  add-codes int2bytes shout32 ;
: triad 3 lightn ;
: 7th   4 lightn ;
: 9th   5 lightn ;

\ now map to musical notes and multiple keys

\ chords
\ Main chord types (maj, min, dim, aug): https://www.schoolofrock.com/resources/keyboard/piano-chords-for-beginners
\ Triads: https://en.wikipedia.org/wiki/Triad_(music)
\ Arpeggio: https://en.wikipedia.org/wiki/Arpeggio
\ Sevenths: https://www.musictheory.net/lessons/45
\ Ninths: https://blog.discmakers.com/2020/12/ninth-chords-add-ambiance/
\ Elevenths: https://blog.discmakers.com/2021/03/eleventh-chords/

variable rt  \ root note
: rt! ( key-id -- key-id ) dup rt c! ;
: rt@ ( -- key-id ) rt c@ ;
variable shp \ chord
: shp! ( shp-id -- ) shp c! ;
: shp@ ( -- shp-id ) shp c@ ;
0 constant majshp
1 constant minshp
2 constant dimshp
3 constant augshp
4 constant sus4shp
5 constant sus2shp
6 constant dom7shp
7 constant maj7shp
8 constant min7shp
9 constant hdim7shp \ half-diminished 7th
10 constant dim7shp  \ fully-diminished 7th

\ major chord shape: root + 4 + 3
: mj   ( root-id -- ) rt! dup 4 + dup 3 + triad majshp  shp! ;  \ ex. c-maj = C E G ;  
: mn   ( root-id -- ) rt! dup 3 + dup 4 + triad minshp  shp! ;
: dm   ( root-id -- ) rt! dup 3 + dup 3 + triad dimshp  shp! ;
: ag   ( root-id -- ) rt! dup 4 + dup 4 + triad augshp  shp! ;
: sus4 ( root-id -- ) rt! dup 5 + dup 2 + triad sus4shp shp! ;
: sus2 ( root-id -- ) rt! dup 2 + dup 5 + triad sus2shp shp! ;
: dom7 ( root-id -- ) rt! dup 4 + dup 3 + dup 3 + 7th dom7shp shp! ;  \ ex. c-maj7 = C E G B ;
: mj7  ( root-id -- ) rt! dup 4 + dup 3 + dup 4 + 7th maj7shp shp! ;  \ ex. c-maj7 = C E G B ;
: mn7  ( root-id -- ) rt! dup 3 + dup 4 + dup 3 + 7th min7shp shp! ;
: hdm7 ( root-id -- ) rt! dup 3 + dup 3 + dup 4 + 7th hdim7shp shp! ;
: dm7  ( root-id -- ) rt! dup 3 + dup 3 + dup 3 + 7th dim7shp shp! ; 

create chord-exec ' mj , ' mn , ' dm , ' ag , ' sus4 , ' sus2 , ' dom7 , ' mj7 , ' mn7 , ' hdm7 , ' dm7 ,  
: runchord ( shp-id -- ) 4 * chord-exec + @ execute ;

\ To transpose the chord, move the shape
: chf ( -- ) rt@ 1 + rt! shp@ runchord ;
: chb ( -- ) rt@ 1 - rt! shp@ runchord ;


\ notes
 8 constant cc
10 constant dd
12 constant ee
13 constant ff
15 constant gg
17 constant aa
19 constant bb
20 constant cc'
: %% 1 - ;
: ## 1 + ;

: kk light-key ;
: mm mj ;
: -+ disable-output 200 ms enable-output ;  \ flash
: hh -+ ; 

\ demo - piano lights board  ---- this forms an EXCELLENT introduction to intermediate forth: memory, data structures, and arrays
\                                 within the powerful context of hardware interfacing/visual programming  (try out with Jai)
\                                 requires Arduino Nano microcontroller with 3iForth.avr, usb cable, and piano lights board 

: kk_ kk 50 ms ;
: l2r32  33 0 do i kk_ loop ;
: r2l32  31 0 do 31 i - kk_ loop ;
: kit32  begin l2r32 r2l32 key? until clear32 ;

create blues 2 c, 4 c, 6 c, 9 c, 11 c, 14 c, 16 c, 18 c, 21 c, 23 c, 26 c, 28 c, 30 c, 
create whites 1 c, 3 c, 5 c, 7 c, 8 c, 10 c, 12 c, 13 c, 15 c, 17 c, 19 c, 20 c, 22 c, 24 c, 25 c, 27 c, 29 c, 31 c, 
13 constant num-blues
19 constant num-whites

: color-l2r  ( array narray -- ) swap c@+ rot 0 do kk_ c@+ loop 2drop ;
: color-r2l  ( array narray -- ) byte go-to-end swap c@- rot 0 do kk_ c@- loop 2drop ;
: blues-l2r   blues num-blues color-l2r ; 
: blues-r2l   blues num-blues color-r2l ;
: blues-kit   begin blues-l2r blues-r2l key? until clear32 ;
: whites-l2r  whites num-whites color-l2r ; 
: whites-r2l  whites num-whites color-r2l ;
: whites-kit  begin whites-l2r whites-r2l key? until clear32 ;
: rtw-cw      begin blues-l2r whites-r2l key? until clear32 ;
: rtw-ccw     begin whites-l2r blues-r2l key? until clear32 ;

: run-til-key  ( xt -- )  begin dup execute key? until drop key drop ;
: lights-demo  ['] kit32 run-til-key 
               ['] blues-kit run-til-key 
               ['] whites-kit run-til-key 
               ['] rtw-cw run-til-key 
               ['] rtw-ccw run-til-key ;



\ ------------------------
\ help
: help-pianolights ( -- ) 
  cr ." help-pianolights"
  cr ." Help for 32-key Piano Lights Board"
  cr ." ----------------------------------"
  cr ."    include 3iForth.fs, 3iSound.fs, 3iShift.fs, 3iSound-4x4.fs, 3iPianoLights.fs"
  cr ."    1 open-com  rx "
  cr ."    init-shiftout    13 start-speaker"
  cr ."    32-on   clear32"
  cr ."    1 light-key   32 light-key  or  kk ( single notes )"
  cr ."    cc dd ee ff gg aa bb cc' ( notes )"
  cr ."    8 12 13 triad  ( triad of notes ) or mm for maj shape"
  cr ."    8 12 15 19  7th  ( four notes ) "
  cr ."    8 12 15 19 22  9th ( five notes ) "
  cr ."    1 mj    1 mn   1 dm   1 ag    1 sus4   1 sus2  ( chords )"
  cr ."    1 dom7  1 mj7  1 mn7  1 hdm7  1 dm7"
  cr ."    -+ or hh  ( flash light )"
  cr ."    chf  chb  ( chord shape semitone transpositions )"
  cr ."    kit32  blues-kit  whites-kit  rtw-cw  rtw-ccw  (demo)"
  cr ."    lights-demo   ( press key to cycle through )"
;

help-pianolights


\ VERSIONS
\ V1 2021-05-11---Tue---05:05 - ADDED: PianoLights.fs note and chord mappings and major, minor, dim, aug, sus4/sus2 chords
\ V2 2021-05-12---Wed---02:22 - ADDED: 7th and 9th chords, single note names, flash, and full help file
\ V3 2021-05-14---Fri---05:27 - ADDED: lights-demo for the piano lights board
\ V4 2021-05-15---Sat---16:56 - ADDED: Arpeggiator feature, controllable from keys or keyboard, with lights!  Integrated with 3iRecord.fs
