( Recording Program for Programmable Digital Synthesizer running 3-instruction Forth on Atmel 328 Arduino Nano )

\ THIS NEEDS FAIR BIT OF WORD TO GET IT RECORDING PROPERLY

: record  ( on1/off0 -- )   \ turns on/off piano serial interface -- when a note is played, piano sends digital note data
    xstore CALL_RECORD xcall ; 
\ : [..] ( n -- idx ) \ plays note from scale array, 1 through 8 is from scale8
\ 	1 - scale + c@ ;

\ * words for PARSING digital piano note data * 
variable numnotes     \ number of entries in score
	: numnotes++ ( -- ) numnotes c@ 1 + numnotes c! ;
	: reset-notes ( -- ) 0 numnotes c! ;
	: numnotes@ numnotes c@ ;
    : del-note  numnotes c@ 1 - numnotes c! ;

\ * DATA STRUCTURE:  note duration rest, each is 2 bytes *
create notes 255 wwords allot  \ holds notes as idx into piano frequency array (later could be frequencies if needed)
create dur   255 wwords allot  \ holds durations
create rest  255 wwords allot  \ holds rest before indexed note
create score 3 dwords allot    \ holds pointers to note, duration, and rests
: init-score  ( -- ) \ notes dur rest
        rest dur notes score !+ !+ !+ drop 
        reset-notes ( 3 0 set-zero ) ;
\ fetches the individual pointers
: *n ( -- n-ptr ) score @ ;
: *d ( -- d-ptr ) score 1 dwords + @ ;
: *r ( -- r-ptr ) score 2 dwords + @ ;
\ save note data
: !n ( n -- )    *n  numnotes@  wwords +    ww! ;
: !d ( dur -- )  *d   numnotes@ wwords +    ww! ;
: !r ( rest -- ) *r   numnotes@ wwords +    ww! ;
\ show note data
: show-notes  cr numnotes@ dup if 0 do *n i wwords +  ww@ . ." .. "  loop cr else drop then ;
: show-dur  cr numnotes@ 0 do *d i wwords +  ww@ . ." " loop cr ;
: show-rest  cr numnotes@ 0 do *r i wwords +  ww@ . ." " loop cr ;

\ setup
: record-on  1 record   init-score   ( 3 0 set-zero ) ;
: record-off 0 record  ( reset-zero ) ;

: [parse-note] ( adr -- adr' val16 )  \ reads byte buffer parsed as little-endian words with low byte first to val16
    c@+ swap      c@+ 8 lshift     rot    or ;
: parse-note ( adr -- idx )  \ 13 fields to a note record
  rcv-buffer
  [parse-note] ( dup   hex. ) !d
  [parse-note] ( dup   hex. ) !r
  [parse-note] ( dup   hex. ) !n
  numnotes++
  drop ;

\ with the number decoded, play it using ., relative to 3 0 set-zero (or the start note of the piano key).  Or could remap to a scale and use ..

\ Simplified receive - key: it is not a realtime read

\ 6-byte key press tuples
: fetch-note-data  ( <serial> -- numbytes )  \ reads from serial buffer into rcv-buffer
    6 read-com ;
: show-note-data
    rcv-buffer 6 show-rcv 2drop ;
: extract-note ( -- note-id ) 
    rcv-buffer 2 wwords + [parse-note] nip ;
    

: get-note ( -- note/stop )  \ read note from serial, 0 means no note, 1-13 notes
    numnotes@ $FF < if fetch-note-data dup if \ show-rcv cr
			  parse-note then 
    else cr ." Memory full.  reset-notes or del-note." cr
	then ;
: get-notes ( N -- ) 0 cr do get-note drop loop ;
: fix-first-note-rest  ( numnotes-first -- ) wwords *r + 5000 swap ww! ;

: all-notes ( -- )  cr numnotes@ >r begin get-note 0= until cr numnotes@ . ." notes logged." cr r> fix-first-note-rest show-notes ;
\ : play-notes-only ( -- ) 
\     numnotes@  dup  if 0 do 
\ 		*n i wwords + ww@    ., 
\ 	loop else drop then ;
\ : play-notes-fixedr ( -- ) 
\     numnotes@  dup  if 0 do 
\ 		R2   \ fixed rest
\ 		*n i wwords + ww@    1 - show-zero + freq lohi 
\ 		*d i wwords + ww@ 30 * lohi
\ 		speaker 
\ 		[tone] 
\ 	loop else drop then ;
: spin ( n -- ) 0 do loop ;
\ : play-notes-fixedd ( -- ) 
\     numnotes@  dup  if 0 do 
\ 		*r i wwords + ww@ 1000 / 1 + 100 * ms 
\ 		*n i wwords + ww@ 1 - show-zero + freq lohi 
\ 		100 lohi  \ fixed duration
\ 		speaker 
\ 		[tone]
\ 	loop else drop then ;
\ : play-notes ( -- ) 
\     numnotes@  dup  if 0 do 
\ 		*r i wwords + ww@ 1000 / 1 + 100 * ms 
\ 		*n i wwords + ww@ 1 - show-zero + freq lohi 
\ 		*d i wwords + ww@ 30 * lohi
\ 		speaker 
\ 		[tone] 
\ 	loop else drop then ;
\ : scale-notes ( -- ) 
\     numnotes@  dup  if 0 do 
\ 		*r i wwords + ww@ 1000 / 1 + 80 * ms 
\ 		*n i wwords + ww@ [..] freq lohi 
\ 		*d i wwords + ww@ 30 * lohi
\ 		speaker 
\ 		[tone]
\ 	loop else drop then ;
\ : scale-play ( -- ) 
\     cr numnotes@  dup  
\ 	if 0 
\ 		do 
\ 			R1
\ 			*n i wwords + ww@ dup >r [..] freq lohi 
\ 			*d i wwords + ww@ 30 * lohi
\ 			speaker
\ 			[tone] 
\ 			key r> over '-' = if  drop else . ." .. " 
\ 			dup '1' = if ." R1 " then '2' = if ." R2 " then then
\ 		loop 
\ 	else drop 
\ 	then ;

: help-record
  cr ." To record on your digital piano: "
  cr ."    record-on"
  cr ."    <play notes>"
  cr ."    get-note   all-notes  play-notes   del-note  reset-notes"
  cr ."    show-notes  show-dur  show-rest"
  cr ."    record-off"
;


\ =====================
\ V1 - 
\ V2 - 
\ V3 - 2021-05-15---Sat---16:57 - Integrated with Arpeggitor on piano board with lights board and keyboard (not 4x4"

